/*****************************************************************************/
/* DUART.c                                                                   */
/*                                                                           */
/* DO NOT ENABLE OPTIMIZER!!!                                                */
/* --------------------------                                                */
/*                                                                           */
/* Driver source for DUART RS-232 card based on MC68681 compatible chip      */
/* http://ide64.come.to                                                      */
/*                                                                           */
/* This module is based upon the public domain swiftlink module written by   */
/* Craig Bruce and adapted to cc65 C compiler by Ullrich von Bassewitz       */
/* http://www.cc65.org                                                       */
/*                                                                           */
/* Ullrich von Bassewitz, 1999                                               */
/* Josef Soucek, 2002                                                        */
/* E-mail: josef.soucek@ct.cz                                                */
/*                                                                           */ 
/* This software is provided 'as-is', without any expressed or implied       */
/* warranty.  In no event will the authors be held liable for any damages    */
/* arising from the use of this software.                                    */
/*                                                                           */
/* Permission is granted to anyone to use this software for any purpose,     */
/* including commercial applications, and to alter it and redistribute it    */
/* freely, subject to the following restrictions:                            */
/*                                                                           */
/* 1. The origin of this software must not be misrepresented; you must not   */
/*    claim that you wrote the original software. If you use this software   */
/*    in a product, an acknowledgment in the product documentation would be  */
/*    appreciated but is not required.                                       */
/* 2. Altered source versions must be plainly marked as such, and must not   */
/*    be misrepresented as being the original software.                      */
/* 3. This notice may not be removed or altered from any source              */
/*    distribution.                                                          */
/*                                                                           */
/*****************************************************************************/

#include <stdio.h>
#include "duart.h"

void NmiHandler (void);
void TryToSend (void);

#define DUART 0xde00
#define CHANNEL  A 

#ifdef CHANNEL

/* DUART channel A */
#define MR1A     DUART     //Mode register, channel A(1)
#define MR2A     DUART     //Mode register, channel A(2)
#define SRA      DUART+1   //Status register, channel A
#define CSRA     DUART+1   //Clock select register, channel A
#define MISR     DUART+2   //@@@@ Masked interrupt status register @@@@ not in MC68681
#define CRA      DUART+2   //Command register, channel A
#define RHRA     DUART+3   //Rx holding register, channel A
#define THRA     DUART+3   //Tx holding register, channel A
#define IPCR     DUART+4   //Input port change register
#define ACR      DUART+4   //Auxiliary control register
#define ISR      DUART+5   //Interrupt status register
#define IMR      DUART+5   //Interrupt mask register
#define CTU      DUART+6   //Counter/timer Upper byte register
#define CTL      DUART+7   //Counter/timer Lower byte register

#define MR1B     DUART+8   //Mode register, channel B(1)
#define MR2B     DUART+8   //Mode register, channel B(2)
#define SRB      DUART+9   //Status register, channel B
#define CSRB     DUART+9   //Clock select register, channel B
#define CRB      DUART+10  //Command rgister, channel B
#define RHRB     DUART+11  //Rx holding register, channel B
#define THRB     DUART+11  //Tx holding register, channel B
#define IVR      DUART+12  //Interrupt vector register
#define IP       DUART+13  //Input port
#define OPCR     DUART+13  //Output port configuration register
#define SCC      DUART+14  //Start counter/timer command
#define SOPBC    DUART+14  //Set output port bit command
#define STC      DUART+15  //Stop counter/timer command
#define COPBC    DUART+15  //Clear output port bits 1 command

#else           /* CHANNEL */

/* DUART channel B */
#define MR1A     DUART+8   //* Mode register
#define MR2A     DUART+8   //* Mode register
#define SRA      DUART+9   //* Status register
#define CSRA     DUART+9   //* Clock select register
#define MISR     DUART+2   //Masked interrupt status register
#define CRA      DUART+10  //* Command register
#define RHRA     DUART+11  //* Rx holding register
#define THRA     DUART+11  //* Tx holding register
#define ACR      DUART+4   //Auxiliary control register
#define IMR      DUART+5   //Interrupt mask register

#define IP       DUART+13  //Input port
#define SOPBC    DUART+14  //Set output port bit command
#define COPBC    DUART+15  //Clear output port bits 1 command

#endif          /* CHANNEL */

static unsigned char DropCnt[]={0,0,0,0}; // Number of bytes lost from rx buffer full
static unsigned char Initialized=0;       // Flag indicating driver is initialized
static unsigned char Stopped=0;           // Flag indicating asserted RTS
static unsigned char Errors=0;            // Number of bytes received in error
static unsigned char RecvFreeCnt=0;       // Number of free bytes in receive buffer
static unsigned char SendFreeCnt=0;       // Number of free bytes in send buffer
static unsigned char RecvHead=0;          // Head of receive buffer
static unsigned char SendHead=0;          // Head of send buffer
static unsigned char RecvTail=0;          // Tail of receive buffer
static unsigned char SendTail=0;          // Tail of send buffer
static unsigned char BaudCode=0;          // Current baud in effect
static unsigned char Tmp_X=0;             // Backup X
static unsigned char Tmp_Y=0;             // Backup Y


static unsigned char RecvBuf[256];
static unsigned char SendBuf[256];

static unsigned char baud_table[]={0x00,  //50
                                   0x11,  //110
                                   0x22,  //134.5
                                   0x44,  //300
                                   0x55,  //600
                                   0x66,  //1200
                                   0x88,  //2400
                                   0x99,  //4800
                                   0xbb,  //9600
                                   0xcc,  //19200
                                   0xcc,  //38400
                                   0x77,  //57600
                                   0x88,  //115200
                                   0xff,  //Not Implemented
                                   0xff,  //Not Implemented
                                   0xff}; //Not Implemented

static unsigned char baud_table2[]={0,0,0,0,0,0,0,0,0,1,0,1,1};

static unsigned char NmiContinue=0x4c;  //JMP instruction for NMI
static unsigned char NmiSave[]={0,0};

unsigned char duart_rs232_init (char hacked)
{
  asm("bit _Initialized");
  asm("bpl @L1");                       // Jump if not initialized
  asm("jsr _duart_rs232_done");         // JSR if initialized
  asm("@L1:");
                                        // Initialize buffers & control
  asm("lda #0");
  asm("sta _RecvHead");
  asm("sta _SendHead");
  asm("sta _RecvTail");
  asm("sta _SendTail");
  asm("sta _Errors");
  asm("sta _Stopped");
  asm("sta _DropCnt");
  asm("sta _DropCnt+1");
  asm("sta _DropCnt+2");
  asm("sta _DropCnt+3");
  asm("lda #255");
  asm("sta _RecvFreeCnt");              // #$ff= free   #$00= full
  asm("sta _SendFreeCnt");

  asm("lda $0318");                     // NMIVec
  asm("ldy $0319");
  asm("sta _NmiSave");
  asm("sty _NmiSave+1");
  asm("lda #<_NmiHandler");
  asm("ldy #>_NmiHandler");
  asm("sta $0318");                     // NMIVec
  asm("sty $0319");

//** set default to 2400-8N1, enable interrupts

  asm("lda #%%00100000");
  asm("sta %w",CRA);                    // Reset Receiver
  asm("lda #%%00110000");
  asm("sta %w",CRA);                    // Reset Transmitter

#ifdef CHANNEL
  asm("lda #%%01010000");               // Set Indicator
#else
  asm("lda #%%10100000");
#endif
  asm("sta %w",OPCR);
  
  asm("lda #%%10110000");               // transmit X=0
  asm("sta %w",CRA);
  asm("lda #%%10010000");               // receiver X=0
  asm("sta %w",CRA);
  asm("lda #%%00000000");               // Bit Rate Set #1
  asm("sta %w",ACR);
  asm("lda #%%10001000");               // X=0 2400
  asm("sta %w",CSRA);
  
  asm("lda #%%00010000");               // Rst MR point
  asm("sta %w",CRA);
  asm("lda #%%01010011");               // 8bit no par + interrupt [bit6] rxready 1=FIFO FULL
  asm("sta %w",MR1A);
  asm("lda #%%00000111");               // 1 stop bit
  asm("sta %w",MR2A);

#ifdef CHANNEL
  asm("lda #%%00000010");               // Enable receive interrupt
#else
  asm("lda #%%00100000");
#endif
  asm("sta %w",IMR);

  asm("lda #%%00000101");               // Enable tx,rx
  asm("sta %w",CRA);

#ifdef CHANNEL
  asm("lda #%%00000101");               // DTR = 1, RTS = 1 Channel A
#else
  asm("lda #%%00001010");               // DTR = 1, RTS = 1 Channel B
#endif
  asm("sta %w",SOPBC);

  asm("lda #6");
  asm("sta _BaudCode");
  
  asm("lda #$ff");
  asm("sta _Initialized");
  return 0;
}

unsigned char duart_rs232_params (unsigned char params, unsigned char parity)
{
  asm("bit _Initialized");
  asm("bmi @L1");                       // Jump if initialized
  asm("jmp @NotInitialized");
  asm("@L1:");

  asm("cmp #0");
  asm("beq @Par_ok");                   // None
  return RS_ERR_BAUD_NOT_AVAIL;         // Parity not implemented

  asm("@Par_ok:");
  asm("lda #%%00010000");               // Rst MR point
  asm("sta %w",CRA);
  asm("lda #%%01010011");               // 8bit no par
  asm("sta %w",MR1A);
  asm("lda #%%00000111");               // 1 stop bit
  asm("sta %w",MR2A);
  
  asm("ldy #1");                        // "params" argument
  asm("lda (sp),y");

  asm("and #$0f");
  asm("sta _BaudCode");
  asm("tay");
  asm("lda _baud_table,y");
  asm("cmp #$ff");
  asm("bne @L5");
  return RS_ERR_BAUD_NOT_AVAIL;

  asm("@L5:");
  asm("sta %w",CSRA);
  asm("lda _baud_table2,y");
  asm("beq @Baud_x0");

  asm("lda #%%10100000");               // transmit X=1
  asm("sta %w",CRA);
  asm("lda #%%10000000");               // receiver X=1
  asm("sta %w",CRA);
  asm("jmp @LA");

  asm("@Baud_x0:");
  asm("lda #%%10110000");               // transmit X=0
  asm("sta %w",CRA);
  asm("lda #%%10010000");               // receiver X=0
  asm("sta %w",CRA);

  asm("@LA:");
  asm("lda #%%00000000");               // Bit Rate Set #1
  asm("sta %w",ACR);
  return 0;
  
  asm("@NotInitialized:");
  return RS_ERR_NOT_INITIALIZED;
}
unsigned char duart_rs232_done (void)
{
  asm("bit _Initialized");
  asm("bpl @L9");               // Jump if not initialized

  asm("lda #0");
  asm("sta %w",IMR);            //Stop Interrupts
  asm("lda #%%00100000");
  asm("sta %w",CRA);            //Reset Receiver        - Flush FIFO, disable receiver
  asm("lda #%%00110000");
  asm("sta %w",CRA);            //Reset Transmitter     - disable receiver
#ifdef CHANNEL
  asm("lda #%%00000101");
#else
  asm("lda #%%00001010");
#endif
  asm("sta %w",COPBC);          //Reset DTR, RTS

  // Restore NMI vector

  asm("lda _NmiSave");
  asm("ldy _NmiSave+1");
  asm("sta $0318");             //NMIVec
  asm("sty $0319");

  asm("@L9:");
  asm("lda #0");
  asm("sta _Initialized");
  return 0;
}
unsigned char duart_rs232_get (char* b)
{
  asm("bit _Initialized");
  asm("bpl @NotInitialized");   // Jump if not initialized

  asm("sta ptr1");
  asm("stx ptr1+1");            // Store pointer to received char

  // Check for bytes to send
  // ldx _SendFreeCnt
  // cpx #$ff
  // beq @L2
  // lda #0
  // jsr TryToSend
  // asm("@L2:");
 
  asm("lda _RecvFreeCnt");
  asm("cmp #$ff");
  asm("bne @L3");

  // No Data in Receive buffer, what about DUART FIFO ?

  asm("lda #0");                // 2
  asm("sta %w",IMR);            // 4     // Stop NMI source of interrupt

  asm("lda %w",SRA);
  asm("and #%%00000001");
  asm("beq @Exit");
  
  asm("lda %w",RHRA);           // 4     // Data reg.
  asm("ldy _RecvTail");         // 4
  asm("sta _RecvBuf,y");        // 5
  asm("inc _RecvTail");         // 6
  asm("dec _RecvFreeCnt");      // 6

#ifdef CHANNEL
  asm("lda #%%00000010");                // Enable receive interrupt
#else
  asm("lda #%%00100000");
#endif
  asm("sta %w",IMR);
  asm("jmp @L4");

  asm("@Exit:");
#ifdef CHANNEL
  asm("lda #%%00000010");               // Enable receive interrupt
#else
  asm("lda #%%00100000");
#endif
  asm("sta %w",IMR);
  return RS_ERR_NO_DATA;                // No characters are available

  asm("@L3:");
  asm("ldx _Stopped");
  asm("beq @L4");
  asm("cmp #63");                       // High space limit for RTS
  asm("bcc @L4");
  asm("lda #0");
  asm("sta _Stopped");
#ifdef CHANNEL
  asm("lda #%%00000101");               // DTR = 1, RTS = 1 Channel A
#else
  asm("lda #%%00001010");               // DTR = 1, RTS = 1 Channel B
#endif
  asm("sta %w",SOPBC);

  asm("@L4:");
  asm("ldx _RecvHead");
  asm("lda _RecvBuf,x");
//  asm("sta $0402");                     // @@@@@@@
  asm("ldy #0");
  asm("sta (ptr1),y");                  //  *b=RecvBuf[RecvHead];

  asm("inc _RecvHead");
  asm("inc _RecvFreeCnt");
  return 0;

  asm("@NotInitialized:");
  return RS_ERR_NOT_INITIALIZED;
}
unsigned char duart_rs232_put (char b)
{
  asm("bit _Initialized");
  asm("bpl @NotInitialized");   // Jump if not initialized
  
  asm("ldx _SendFreeCnt");
  asm("cpx #$ff");
  asm("beq @L2");               // Jump if send buffer free

  asm("pha");
  asm("lda #$00");
  asm("jsr _TryToSend");
  asm("pla");

  asm("@L2:");
  asm("ldx _SendFreeCnt");
  asm("bne @L3");               // Jump if send buffer not full
  return RS_ERR_OVERFLOW;       // Send buffer is full

  // Put byte into send buffer & send

  asm("@L3:");
  asm("ldx _SendTail");
  asm("sta _SendBuf,x");
  asm("inc _SendTail");
  asm("dec _SendFreeCnt");
  asm("lda #$ff");              // Try hard
  asm("jsr _TryToSend");
  return 0;

  asm("@NotInitialized:");
  return RS_ERR_NOT_INITIALIZED;
}
void TryToSend(void)
{
  asm("sta tmp1");              // Remember tryHard flag
  asm("@L0:");
  asm("lda _SendFreeCnt");
  asm("cmp #$ff");
  asm("beq @L3");               // Exit if buffer free
  asm("@L1:");
  asm("lda _Stopped");          // Check for flow stopped
  asm("bne @L3");               // Exit if stopped

  asm("lda %w",SRA);
  asm("and #%%00000100");       // (TxRDY)
  asm("bne @L4");
  asm("bit tmp1");              // Keep trying if must try hard
  asm("bmi @L0");
  asm("@L3:");
  asm("rts");
                                // send byte and try again
  asm("@L4:");
  asm("ldx _SendHead");
  asm("lda _SendBuf,x");
//  asm("sta $0403");                     // @@@@@@@
  asm("sta %w",THRA);
  asm("inc _SendHead");
  asm("inc _SendFreeCnt");
  asm("jmp @L0");
} 
unsigned char duart_rs232_pause (void)
{
  asm("inc $d020");
  return 0;
}
unsigned char duart_rs232_unpause (void)
{
  asm("inc $d020");
  return 0;
}
unsigned char duart_rs232_status (unsigned char* status, unsigned char* errors)
{
  *status=*((unsigned char *)SRA);
  *errors=Errors;
  return 0;
}
void NmiHandler(void)
{
  asm("pha");                   // 3

  asm("lda %w",ISR);            // 4     // MISR is not available in MC68681
#ifdef CHANNEL
  asm("and #%%00000010");       // 2     // Receive interrupt RxRDY/FIFO full
#else
  asm("and #%%00100000");
#endif
  asm("beq @L9");               // 2     // Exit MNI was from another source

  asm("lda #0");                // 2
  asm("sta %w",IMR);            // 4      // Stop NMI source of interrupt

  asm("lda %w",RHRA);           // 4      // Data reg. (get as soon as possible)

  asm("cld");                   // 2
  asm("stx _Tmp_X");            // 4
  asm("sty _Tmp_Y");            // 4
  asm("jmp @NMICont");

  asm("@NMILoop:");
  asm("lda %w",RHRA);           // 4      // Data reg.

  asm("@NMICont:");
  asm("ldy _RecvTail");         // 4
  asm("ldx _RecvFreeCnt");      // 4
  asm("beq @L3");               // 2    // Buffer full
  asm("sta _RecvBuf,y");        // 5
  asm("inc _RecvTail");         // 6
  asm("dec _RecvFreeCnt");      // 6
  asm("cpx #33");               // 2    // Low space limit for RTS
  asm("bcs @Nmi_Cont");

#ifdef CHANNEL
  asm("lda #%%00000101");                 // DTR = 0, RTS = 0 Channel A
#else
  asm("lda #%%00001010");                 // DTR = 0, RTS = 0 Channel B
#endif
  asm("sta %w",COPBC);
  asm("sta _Stopped");

  asm("@Nmi_Cont:");
  asm("lda %w",SRA);
  asm("and #%%00000001");
  asm("bne @NMILoop");

  asm("@NmiExit:");
  asm("ldx _Tmp_X");            // 4
  asm("ldy _Tmp_Y");            // 4

#ifdef CHANNEL
  asm("lda #%%00000010");               // Enable receive interrupt
#else
  asm("lda #%%00100000");
#endif
  asm("sta %w",IMR);

  asm("pla");
  asm("rti");

// ###########
//  asm("lda %w",SRA);
//  asm("and #%%01110000");       // 4      // Check Framing, parity, overrun error
//  asm("beq @L1");               // 3
//  asm("inc _Errors");
//  asm("@L1:");
// ###########

  asm("@L3:");
  asm("inc $d021");                     // @@@@@@@
  asm("inc _DropCnt");
  asm("bne @L4");
  asm("inc _DropCnt+1");
  asm("bne @L4");
  asm("inc _DropCnt+2");
  asm("bne @L4");
  asm("inc _DropCnt+3");
  asm("@L4:");
  asm("jmp @NmiExit");
  
  asm("@L9:");                           // NMI was from another source than DUART
  asm("pla");                   // 4 
  asm("jmp %v",NmiContinue);
}

